#define BOOST_TEST_MODULE "testcase for xmlhelper functions"
#define BOOST_TEST_DYN_LINK
#include "common.hpp"

#include <boost/test/unit_test.hpp>

#include <cmath>
#include <glibmm/ustring.h>
#include <harkio2/xmlhelper/xmlhelper.hpp>
#include <harkio2/Position.hpp>
#include <harkio2/AnalyzeError.hpp>
#include <harkio2/FileType.hpp>
#include <harkio2/harkconfig/nodeparsers.hpp>
#include <harkio2/harkconfig/Source.hpp>
#include <harkio2/harkconfig/File.hpp>
#include <harkio2/harkconfig/Mic.hpp>
#include <harkio2/harkconfig/NoiseSource.hpp>
#include <harkio2/harkconfig/HarkConfigFile.hpp>
#include <harkio2/harkconfig/ImpulseResponseListFile.hpp>

#include "rangecheck.hpp"

using namespace std;
using namespace rangecheck;
namespace xmlhelper = harkio2::xmlhelper;
namespace harkconfig = harkio2::harkconfig;

const char* xmlPosition1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position unit="none" type="unitvector">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
)xml";

const char* xmlPosition2 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position angle="radian" type="polar">
    <Theta>0.78539816339744828</Theta>
    <Phi>0</Phi>
</Position>
)xml";

const char* xmlPosition3 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position unit="meter" type="cartesian">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
)xml";

const char* xmlPosition4 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position unit="meter" angle="radian" type="polar">
    <Theta>0.78539816339744828</Theta>
    <Phi>0</Phi>
    <R>1.5</R>
</Position>
)xml";

const char* xmlPosition5 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position unit="meter" type="unitvector">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
    <R>1.5</R>
</Position>
)xml";

BOOST_AUTO_TEST_CASE(ParsePosition_test)
{
    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlPosition1), strlen(xmlPosition1));

    BOOST_CHECK(reader1.read());
    BOOST_CHECK(xmlhelper::IsElement(reader1, "Position"));

    harkio2::Position<float> pos = harkconfig::ParsePosition(reader1);

    BOOST_CHECK(xmlhelper::IsEndElement(reader1, "Position"));

    // position check
    BOOST_CHECK(pos.IsDirection());
    BOOST_CHECK(CheckRange<float>(pos.GetUnitVector(), Normalize(harkio2::Vector3<float>(1.0, 1.0, 0)), 0.01));

    xmlpp::TextReader reader2(reinterpret_cast<const unsigned char*>(xmlPosition2), strlen(xmlPosition2));
    reader2.read();
    pos = harkconfig::ParsePosition(reader2);
    BOOST_CHECK(pos.IsDirection());
    BOOST_CHECK(CheckRange<float>(pos.GetUnitVector(), Normalize(harkio2::Vector3<float>(1.0, 1.0, 0)), 0.01));

    xmlpp::TextReader reader3(reinterpret_cast<const unsigned char*>(xmlPosition3), strlen(xmlPosition3));
    reader3.read();
    pos = harkconfig::ParsePosition(reader3);
    BOOST_CHECK(pos.IsPosition());
    BOOST_CHECK(CheckRange<float>(pos.GetUnitVector(), Normalize(harkio2::Vector3<float>(1.0, 1.0, 0)), 0.01));
    BOOST_CHECK(CheckRange<float>(pos.GetLength(), sqrt(2.0), 0.01));

    xmlpp::TextReader reader4(reinterpret_cast<const unsigned char*>(xmlPosition4), strlen(xmlPosition4));
    reader4.read();
    pos = harkconfig::ParsePosition(reader4);
    BOOST_CHECK(pos.IsPosition());
    BOOST_CHECK(CheckRange<float>(pos.GetUnitVector(), Normalize(harkio2::Vector3<float>(1.0, 1.0, 0)), 0.01));
    BOOST_CHECK(CheckRange<float>(pos.GetLength(), 1.5, 0.01));

    xmlpp::TextReader reader5(reinterpret_cast<const unsigned char*>(xmlPosition5), strlen(xmlPosition5));
    reader5.read();
    pos = harkconfig::ParsePosition(reader5);
    BOOST_CHECK(pos.IsPosition());
    BOOST_CHECK(CheckRange<float>(pos.GetUnitVector(), Normalize(harkio2::Vector3<float>(1.0, 1.0, 0)), 0.01));
    BOOST_CHECK(CheckRange<float>(pos.GetLength(), 1.5, 0.01));

}

// erroneous
const char* xmlPosition6 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position unit="hoge" type="unitvector">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
)xml";

const char* xmlPosition7 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position type="unitvector">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
)xml";

// erroneous
const char* xmlPosition8 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position type="hoge">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
)xml";

// erroneous
const char* xmlPosition9 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Position>
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
)xml";

BOOST_AUTO_TEST_CASE(ParsePosition_test2)
{
    harkio2::Position<float> pos;

    xmlpp::TextReader reader6(reinterpret_cast<const unsigned char*>(xmlPosition6), strlen(xmlPosition6));
    xmlpp::TextReader reader7(reinterpret_cast<const unsigned char*>(xmlPosition7), strlen(xmlPosition7));
    xmlpp::TextReader reader8(reinterpret_cast<const unsigned char*>(xmlPosition8), strlen(xmlPosition8));
    xmlpp::TextReader reader9(reinterpret_cast<const unsigned char*>(xmlPosition9), strlen(xmlPosition9));

    reader6.read();
    reader7.read();
    reader8.read();
    reader9.read();

    BOOST_CHECK_THROW((pos = harkconfig::ParsePosition(reader6)), harkio2::AnalyzeError);

    BOOST_CHECK_NO_THROW((pos = harkconfig::ParsePosition(reader7)));
    BOOST_CHECK(pos.IsDirection());

    BOOST_CHECK_THROW((pos = harkconfig::ParsePosition(reader8)), harkio2::AnalyzeError);

    BOOST_CHECK_THROW((pos = harkconfig::ParsePosition(reader9)), harkio2::AnalyzeError);
}

const char* xmlMicIndices1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<MicIndices size="2">
    <MicIndex>10</MicIndex>
    <MicIndex>20</MicIndex>
</MicIndices>
)xml";

const char* xmlMicIndices2 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<MicIndices>
    <MicIndex>10</MicIndex>
    <MicIndex>20</MicIndex>
</MicIndices>
)xml";

// erroneous
const char* xmlMicIndices3 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<MicIndices>
    <MicIndex>10</MicIndex>
    <MicIndex>20</MicIndex>
    <Comment>abc</Comment>
</MicIndices>
)xml";

BOOST_AUTO_TEST_CASE(ParseMicIndices_test)
{
    std::vector<unsigned int> indices;

    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlMicIndices1), strlen(xmlMicIndices1));
    xmlpp::TextReader reader2(reinterpret_cast<const unsigned char*>(xmlMicIndices2), strlen(xmlMicIndices2));
    xmlpp::TextReader reader3(reinterpret_cast<const unsigned char*>(xmlMicIndices3), strlen(xmlMicIndices3));

    reader1.read();
    reader2.read();
    reader3.read();

    BOOST_CHECK_NO_THROW(indices = harkconfig::ParseMicIndices(reader1));
    BOOST_CHECK_EQUAL(indices.size(), 2);
    if (indices.size() == 2)
    {
        BOOST_CHECK_EQUAL(indices[0], 10);
        BOOST_CHECK_EQUAL(indices[1], 20);
    }

    BOOST_CHECK_NO_THROW(indices = harkconfig::ParseMicIndices(reader2));
    BOOST_CHECK_EQUAL(indices.size(), 2);
    if (indices.size() == 2)
    {
        BOOST_CHECK_EQUAL(indices[0], 10);
        BOOST_CHECK_EQUAL(indices[1], 20);
    }

    BOOST_CHECK_THROW(harkconfig::ParseMicIndices(reader3), harkio2::AnalyzeError);
}

const char* xmlFile1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<File>
    <Name>foo.bar</Name>
    <MicIndices>
        <MicIndex>123</MicIndex>
    </MicIndices>
    <Comment>file comment</Comment>
</File>
)xml";

BOOST_AUTO_TEST_CASE(ParseFile1_test)
{
    harkconfig::File1 file;

    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlFile1), strlen(xmlFile1));

    reader1.read();

    BOOST_CHECK_NO_THROW(file = harkconfig::ParseFile1(reader1));
    BOOST_CHECK_EQUAL(file.GetName(), "foo.bar");
    BOOST_CHECK_EQUAL(file.GetMicIndices().size(), 1);
    if (file.GetMicIndices().size() == 1)
    {
        BOOST_CHECK_EQUAL(file.GetMicIndices()[0], 123);
    }
    BOOST_CHECK_EQUAL(file.GetComment(), "file comment");
}

const char* xmlFiles1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Files>
    <File>
        <Name>foo</Name>
        <MicIndices>
            <MicIndex>123</MicIndex>
        </MicIndices>
        <Comment>file foo</Comment>
    </File>
    <File>
        <Name>bar</Name>
        <MicIndices>
            <MicIndex>456</MicIndex>
        </MicIndices>
        <Comment>file bar</Comment>
    </File>
</Files>
)xml";

BOOST_AUTO_TEST_CASE(ParseFiles1_test)
{
    vector<harkconfig::File1> files;

    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlFiles1), strlen(xmlFiles1));

    reader1.read();

    BOOST_CHECK_NO_THROW(files = harkconfig::ParseFiles1(reader1));
    BOOST_CHECK_EQUAL(files.size(), 2);
    if (files.size() == 2)
    {
        BOOST_CHECK_EQUAL(files[0].GetName(), "foo");
        BOOST_CHECK_EQUAL(files[0].GetMicIndices().size(), 1);
        if (files[0].GetMicIndices().size() == 1)
        {
            BOOST_CHECK_EQUAL(files[0].GetMicIndices()[0], 123);
        }

        BOOST_CHECK_EQUAL(files[1].GetName(), "bar");
        BOOST_CHECK_EQUAL(files[1].GetMicIndices().size(), 1);
        if (files[1].GetMicIndices().size() == 1)
        {
            BOOST_CHECK_EQUAL(files[1].GetMicIndices()[0], 456);
        }

    }
}

const char* xmlSource1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Source>
<Position unit="none" type="unitvector">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
<Files>
    <File>
        <Name>foo</Name>
        <MicIndices>
            <MicIndex>123</MicIndex>
        </MicIndices>
        <Comment>file foo</Comment>
    </File>
    <File>
        <Name>bar</Name>
        <MicIndices>
            <MicIndex>456</MicIndex>
        </MicIndices>
        <Comment>file bar</Comment>
    </File>
</Files>
<CutStart>1</CutStart>
<CutEnd>2</CutEnd>
<Comment>source comment</Comment>
</Source>
)xml";

BOOST_AUTO_TEST_CASE(ParseSource2_test)
{
    harkconfig::Source2 source;

    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlSource1), strlen(xmlSource1));

    reader1.read();

    BOOST_CHECK_NO_THROW(source = harkconfig::ParseSource2(reader1));

    BOOST_CHECK(source.GetPosition().IsDirection());
    BOOST_CHECK(CheckRange<float>(source.GetPosition().GetUnitVector(), Normalize(harkio2::Vector3<float>(1, 1, 0)), 0.01));
    BOOST_CHECK_EQUAL(source.GetFiles().size(), 2);
    if (source.GetFiles().size() == 2)
    {
        BOOST_CHECK_EQUAL(source.GetFiles()[0].GetName(), "foo");
        BOOST_CHECK_EQUAL(source.GetFiles()[0].GetMicIndices().size(), 1);
        if (source.GetFiles()[0].GetMicIndices().size() == 1)
        {
            BOOST_CHECK_EQUAL(source.GetFiles()[0].GetMicIndices()[0], 123);
        }
    }

    BOOST_CHECK_EQUAL(source.GetCutStart(), 1);
    BOOST_CHECK_EQUAL(source.GetCutEnd(), 2);
    BOOST_CHECK_EQUAL(source.GetComment(), "source comment");
}

const char* xmlSources1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Sources>
<Source>
<Position unit="none" type="unitvector">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
<Files>
    <File>
        <Name>foo</Name>
        <MicIndices>
            <MicIndex>123</MicIndex>
        </MicIndices>
        <Comment>file foo</Comment>
    </File>
</Files>
<CutStart>1</CutStart>
<CutEnd>2</CutEnd>
<Comment>source comment</Comment>
</Source>
</Sources>
)xml";

BOOST_AUTO_TEST_CASE(ParseSources2_test)
{
    vector<harkconfig::Source2> sources;

    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlSources1), strlen(xmlSources1));

    reader1.read();

    BOOST_CHECK_NO_THROW(sources = harkconfig::ParseSources2(reader1));

    BOOST_CHECK_EQUAL(sources.size(), 1);
    if (sources.size() == 1)
    {
        harkconfig::Source2& source = sources[0];
        BOOST_CHECK(source.GetPosition().IsDirection());
        BOOST_CHECK(CheckRange<float>(source.GetPosition().GetUnitVector(), Normalize(harkio2::Vector3<float>(1, 1, 0)), 0.01));
        BOOST_CHECK_EQUAL(source.GetFiles().size(), 1);
        if (source.GetFiles().size() == 1)
        {
            BOOST_CHECK_EQUAL(source.GetFiles()[0].GetName(), "foo");
            BOOST_CHECK_EQUAL(source.GetFiles()[0].GetMicIndices().size(), 1);
            if (source.GetFiles()[0].GetMicIndices().size() == 1)
            {
                BOOST_CHECK_EQUAL(source.GetFiles()[0].GetMicIndices()[0], 123);
            }
        }

        BOOST_CHECK_EQUAL(source.GetCutStart(), 1);
        BOOST_CHECK_EQUAL(source.GetCutEnd(), 2);
        BOOST_CHECK_EQUAL(source.GetComment(), "source comment");
    }
}

const char* xmlMic1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Mic>
<MicIndex>123</MicIndex>
<Position unit="meter" type="unitvector">
    <X>1.1</X>
    <Y>2.2</Y>
    <Z>3.3</Z>
    <R>4.4</R>
</Position>
<Comment>mic comment</Comment>
</Mic>
)xml";

BOOST_AUTO_TEST_CASE(ParseMic1_test)
{
    harkconfig::Mic1 mic;
    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlMic1), strlen(xmlMic1));
    reader1.read();

    BOOST_CHECK_NO_THROW(mic = harkconfig::ParseMic1(reader1));
    BOOST_CHECK_EQUAL(mic.GetMicIndex(), 123);
    BOOST_CHECK(mic.GetPosition().IsPosition());
    BOOST_CHECK(rangecheck::CheckRange<float>(mic.GetPosition().GetLength(), 4.4, 0.01));
    BOOST_CHECK_EQUAL(mic.GetComment(), u8"mic comment");
}

const char* xmlMics1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<Mics size="1"><Mic>
    <MicIndex>456</MicIndex>
    <Position unit="meter" type="unitvector">
        <X>1.1</X>
        <Y>2.2</Y>
        <Z>3.3</Z>
        <R>8.8</R>
    </Position>
    <Comment>mic comment</Comment>
</Mic></Mics>
)xml";

BOOST_AUTO_TEST_CASE(ParseMics1_test)
{
    std::vector<harkconfig::Mic1> mics;
    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlMics1), strlen(xmlMics1));
    reader1.read();

    BOOST_CHECK_NO_THROW(mics = harkconfig::ParseMics1(reader1));
    BOOST_CHECK_EQUAL(mics.size(), 1);
    if (mics.size() == 1)
    {
        BOOST_CHECK_EQUAL(mics[0].GetMicIndex(), 456);
        BOOST_CHECK(mics[0].GetPosition().IsPosition());
        BOOST_CHECK(rangecheck::CheckRange<float>(mics[0].GetPosition().GetLength(), 8.8, 0.01));
        BOOST_CHECK_EQUAL(mics[0].GetComment(), u8"mic comment");
    }
}

const char* xmlNoiseSource1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<NoiseSource>
<NoiseIndex>123</NoiseIndex>
<Position unit="none" type="unitvector">
    <X>1.1</X>
    <Y>2.2</Y>
    <Z>3.3</Z>
</Position>
<Comment>noise source comment</Comment>
</NoiseSource>
)xml";

BOOST_AUTO_TEST_CASE(ParseNoiseSource1_test)
{
    harkconfig::NoiseSource1 noiseSource;
    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlNoiseSource1), strlen(xmlNoiseSource1));
    reader1.read();

    BOOST_CHECK_NO_THROW(noiseSource = harkconfig::ParseNoiseSource1(reader1));
    BOOST_CHECK_EQUAL(noiseSource.GetNoiseIndex(), 123);
    BOOST_CHECK(noiseSource.GetPosition().IsDirection());
    BOOST_CHECK_EQUAL(noiseSource.GetComment(), u8"noise source comment");
}

const char* xmlNoiseSources1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<NoiseSources size="1"><NoiseSource>
    <NoiseIndex>456</NoiseIndex>
    <Position unit="none" type="unitvector">
        <X>1.1</X>
        <Y>2.2</Y>
        <Z>3.3</Z>
    </Position>
    <Comment>noise source comment</Comment>
</NoiseSource></NoiseSources>
)xml";

BOOST_AUTO_TEST_CASE(ParseNoiseSources1_test)
{
    std::vector<harkconfig::NoiseSource1> noiseSources;
    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlNoiseSources1), strlen(xmlNoiseSources1));
    reader1.read();

    BOOST_CHECK_NO_THROW(noiseSources = harkconfig::ParseNoiseSources1(reader1));
    BOOST_CHECK_EQUAL(noiseSources.size(), 1);
    if (noiseSources.size() == 1)
    {
        BOOST_CHECK_EQUAL(noiseSources[0].GetNoiseIndex(), 456);
        BOOST_CHECK(noiseSources[0].GetPosition().IsDirection());
        BOOST_CHECK_EQUAL(noiseSources[0].GetComment(), u8"noise source comment");
    }
}

const char* xmlImpulseResponseListFile2_0_1 =
R"xml(<?xml version="1.0" encoding="UTF-8"?>
<hark_config id="ImpulseResponseList" major="2" minor="0">
<NumMic>1</NumMic>
<SourceFileType>float</SourceFileType>
<Comment>irl</Comment>
<Sources>
<Source>
<Position unit="none" type="unitvector">
    <X>1.0</X>
    <Y>1.0</Y>
    <Z>0.0</Z>
</Position>
<Files>
    <File>
        <Name>foo</Name>
        <MicIndices>
            <MicIndex>123</MicIndex>
        </MicIndices>
        <Comment>file foo</Comment>
    </File>
</Files>
<CutStart>1</CutStart>
<CutEnd>2</CutEnd>
<Comment>source comment</Comment>
</Source>
</Sources>
</hark_config>
)xml";

BOOST_AUTO_TEST_CASE(ParseImpulseResponseListFile2_0_test)
{
    boost::shared_ptr<harkconfig::ImpulseResponseListFile2_0> ir;

    xmlpp::TextReader reader1(reinterpret_cast<const unsigned char*>(xmlImpulseResponseListFile2_0_1), strlen(xmlImpulseResponseListFile2_0_1));

    reader1.read();

    BOOST_CHECK_NO_THROW(ir = harkconfig::ParseImpulseResponseListFile2_0(reader1));

    BOOST_CHECK_EQUAL(ir->GetNumMic(), 1);
    BOOST_CHECK_EQUAL(ir->GetSourceFileType(), harkio2::FileType::FLOAT);
    BOOST_CHECK_EQUAL(ir->GetComment(), "irl");

    BOOST_CHECK_EQUAL(ir->GetSources().size(), 1);
}

